# -*- coding: utf-8 -*-

# Copyright (C) 2025 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Ansys SCADE Power Scripts: Create default notes for Scade models."""

import argparse

from ansys.scade.apitools import declare_project

# isort: split
# must be imported after ansys.scade.apitools
from ansys.scade.ps.default_notes.default_notes import main as default_notes_main, tool


def main():
    """Implement ``ansys.scade.ps.default_notes.__main__:main`` packages's script."""
    parser = argparse.ArgumentParser(description=tool)
    parser.add_argument(
        '-p',
        '--projects',
        metavar='<project>',
        help='SCADE Suite projects',
        nargs='+',
        required=True,
    )
    options = parser.parse_args()

    assert declare_project  # nosec B101  # declare_project must be defined on Windows
    for project in options.projects:
        declare_project(project)
    code = default_notes_main()
    exit(code)


if __name__ == '__main__':
    # run with python.exe -m ansys.scade.ps.default_notes
    main()



# Copyright (C) 2025 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Set of utilities."""

from functools import reduce
from re import compile
from typing import List, Sequence, Tuple, Union

from scade.model.suite import NamedType, Structure, Table, Type

LiteralTree = Union[List, str]
"""
Annotation type for scenario values: hierarchy of literal values.
"""


CounterTree = Union[List, int]
"""
Annotation type for sustain counters associated to a scenario value: hierarchy of integers.
"""

# annotation types for utilities common to literal and counter trees
Leaf = Union[str, int]
Tree = Union[LiteralTree, CounterTree]
SubPath = Sequence[Union[str, int]]


def get_type_definition(type_: Type) -> Type:
    """Return the leaf type of a type."""
    if isinstance(type_, NamedType) and not type_.is_predefined():
        return get_type_definition(type_.type)

    return type_


def resolve_path(tree: list, type_: Type, path: SubPath) -> Tuple[list, int, Type]:
    """Return the list containing the last path element, its index and type."""
    subtree = tree
    for elem in path:
        tree = subtree
        # skip aliases
        type_ = get_type_definition(type_)
        # following assert failure(s) implies an error in the scenario
        if isinstance(type_, Table):
            # assert len(subtree) == type_.size
            type_ = type_.type
            # elem must be an integer
            index = int(elem)
        elif isinstance(type_, Structure):
            # assert len(subtree) == len(type_.elements)
            for index, field in enumerate(type_.elements):
                # optimization: add a cache in the type?
                if field.name == elem:
                    index = field.element_range
                    break
            else:
                # not existing field
                raise ValueError(f'{type_.get_full_path()}: unknown {elem} field')
            type_ = field.type
        else:
            raise ValueError(f'{type_.get_full_path()}: value mismatch')
        subtree = tree[index]

    return tree, index, type_


def adjust_value(value: LiteralTree, type_: Type) -> LiteralTree:
    """Replace ``?`` placeholders for tables/structures by trees, bypassing aliases."""
    type_ = get_type_definition(type_)
    if value == '?':
        return get_default_value(type_, '?')  # type: ignore
    if isinstance(type_, Table):
        if not isinstance(value, list) or len(value) != type_.size:
            raise ValueError(f'{type_.get_full_path()}: value mismatch')
        return [adjust_value(cell, type_.type) for cell in value]
    elif isinstance(type_, Structure):
        if not isinstance(value, list) or len(value) != len(type_.elements):
            raise ValueError(f'{type_.get_full_path()}: value mismatch')
        return [
            adjust_value(subvalue, element.type)
            for subvalue, element in zip(value, type_.elements, strict=True)
        ]
    else:
        return value


def patch_tree(
    tree: list, value: LiteralTree, type_: Type, path: Sequence[Union[str, int]], needadjust=False
):
    """Update a value of a tree."""
    tree, index, type_ = resolve_path(tree, type_, path)
    if needadjust:
        tree[index] = adjust_value(value, type_)
    else:
        tree[index] = value


def get_tree_width(tree: CounterTree) -> int:
    """Return the number of elements different from 0 of the flattened tree."""
    if isinstance(tree, list):
        width = reduce(lambda x, y: x + y, [get_tree_width(elem) for elem in tree])
    else:
        width = 1 if tree > 0 else 0
    return width


def patch_sustain(tree: list, sustain: int, type_: Type, path: SubPath) -> int:
    """Update the sustain value of a tree."""
    tree, index, type_ = resolve_path(tree, type_, path)
    old_width = get_tree_width(tree[index])
    tree[index] = get_default_value(type_, sustain)
    new_width = get_type_width(type_) if sustain > 0 else 0
    return new_width - old_width


def get_default_value(type_: Type, fill: object = '') -> Union[List, object]:
    """Return a tree corresponding to a type."""
    # skip aliases
    type_ = get_type_definition(type_)
    if isinstance(type_, Table):
        value = get_default_value(type_.type, fill)
        return [value] * type_.size
    elif isinstance(type_, Structure):
        return [get_default_value(field.type, fill) for field in type_.elements]
    else:
        return fill


_re_path = compile(r'([^\[\.]*)(.*)?')


def split_path(path: str) -> Tuple[str, Sequence[Union[str, int]]]:
    """
    Return the path identifying the semantic element and its subpath as a list: index and/or fields.

    ``path`` is expected to be a sensor or local variable, with optional subpath.
    """
    match = _re_path.match(path)
    assert match is not None  # nosec B101  # addresses linter
    var_path, sub_paths = match.groups()
    if sub_paths:
        items = sub_paths.replace('[', '.').replace(']', '').strip('.').split('.')
        items = [int(i) if i.isdecimal() else i for i in items]
    else:
        items = []

    return var_path, items


def splitex(text: str, separators: str) -> List[str]:
    """Return the list of tokens, including the separators."""
    # use `tok` instead of `token` so that bandit does not raise B105
    tok = ''
    result = []
    quote = False
    for c in text:
        if not quote and c in separators:
            tok = tok.strip()
            if tok != '':
                result.append(tok)
            tok = ''
            result.append(c)
        else:
            tok += c
            if c == "'":
                quote = not quote
    tok = tok.strip()
    if tok != '':
        result.append(tok)

    return result


def value_to_tree(value: str) -> LiteralTree:
    """Provide a tree representation of a structured value."""
    # use `tok[s]` instead of `token[s]` so that bandit does not raise B105
    toks = splitex(value, '()[],')
    tree = []
    current = []
    for tok in toks:
        if tok == ')' or tok == ']':
            pop = tree.pop()
            pop.append(current)
            current = pop
        elif tok == ',':
            pass
        elif tok == '(' or tok == '[':
            tree.append(current)
            current = []
        else:
            current.append(tok)

    return current[0] if current else ''


def reduce_value(value: LiteralTree, reference: LiteralTree) -> LiteralTree:
    """Remove from value parts identical to reference."""
    # value and reference must have the same tree structure
    if not isinstance(value, list):
        # leaf
        return '' if value == reference else value

    result = []
    empty = True
    dontcare = True
    for v, r in zip(value, reference):
        red = reduce_value(v, r)
        if red != '':
            empty = False
        if red != '?':
            dontcare = False
        result.append(red)

    return '' if empty else '?' if dontcare else result


def tree_to_str(value: Union[list, object]) -> str:
    """Serialize a tree to a string."""
    if isinstance(value, list):
        return '(' + ','.join([tree_to_str(v) for v in value]) + ')'
    else:
        # leaf
        return str(value)


# cache
_type_widths = {}


def get_type_width(type_: Type) -> int:
    """Return the number of leaves of the type tree."""
    # use a lazy cache to minimize the computations
    global _type_widths
    width = _type_widths.get(type_)
    if not width:
        if isinstance(type_, Table):
            width = type_.size * get_type_width(type_.type)
        elif isinstance(type_, Structure):
            width = reduce(
                lambda x, y: x + y, [get_type_width(field.type) for field in type_.elements]
            )
        elif isinstance(type_, NamedType) and not type_.is_predefined():
            width = get_type_width(type_.type)
        else:
            width = 1
        _type_widths[type_] = width
    return width


def apply_sustain(value: list, sustain: list) -> int:
    """
    Return the number of obsolete checks when removing one to positive sustain values.

    Set a corresponding value to '?' when its counter becomes null
    """
    count = 0
    # assert len(value) == len(sustain)
    for i, element in enumerate(sustain):
        if isinstance(element, list):
            count += apply_sustain(value[i], sustain[i])
        elif element >= 0:
            if element == 0:
                value[i] = '?'
                count += 1
            sustain[i] -= 1

    return count



# Copyright (C) 2025 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Python library for Scenario Optimizer.

DO NOT EDIT GENERATED BLOCKS, delimited by ``{{`` and ``}}`` markers.
"""

#%% begin

#%% import

from copy import deepcopy
from pathlib import Path
from typing import List, Dict, Optional, Union, Sequence as Seq

# aliases used in ecore references
import scade.model.suite as suite
from scade.model.suite import Object as ScObject, Model as ScModel, ConstVar as ScConstVar

from ansys.scade.ps.optimize_sss.scenparser import SSSParser
from ansys.scade.ps.optimize_sss.scutils import (
    CounterTree,
    LiteralTree,
    split_path,
    value_to_tree,
    patch_tree,
    get_default_value,
    adjust_value,
    get_type_width,
    patch_sustain,
    apply_sustain,
)

# ---------------------------------------------------------------------------
# model
# ---------------------------------------------------------------------------

#%% types

'''
# unused declaration
#{{type(78)
class AliasTable:
    pass
#}}type

#{{type(159)
class SubPath:
    pass
#}}type

#{{type(148)
class ValueTree:
    pass
#}}type
'''


AliasTable = dict
SubPath = Union[Seq[Union[str, int]], None]


class ValueTreeImpl:
    """
    Value as a tree structure.

    Each node is either a literal value, as a string, or a list of nodes.

    .. Notes::

      * `'?`'` represents any sub-tree
      * `''` represents an error in the scenario or an uninitialized value
    """

    def __init__(self, value: str = ''):
        self.tree: LiteralTree = value_to_tree(value)
        # the string value may not be complete with respect of the type
        # for example, '?' instead of (1, 2, 3)
        self.needadjust = '?' in value

    @property
    def scalar(self) -> bool:
        return not isinstance(self.tree, list)

    def clone(self) -> 'ValueTreeImpl':
        clone = ValueTreeImpl()
        clone.tree = deepcopy(self.tree)
        # clone.tree = self.value if self.scalar else self.value.copy()
        clone.needadjust = self.needadjust
        return clone

    def patch(self, value: 'ValueTreeImpl', io: 'IO', path: SubPath):
        # Update a value or sub-value of a variable.
        assert io.constvar is not None  # nosec B101  # addresses linter
        patch = value.tree.copy() if isinstance(value.tree, list) else value.tree
        if path:
            assert isinstance(self.tree, list)  # nosec B101  # addresses linter
            patch_tree(self.tree, patch, io.constvar.type, path, needadjust=value.needadjust)
        else:
            if value.needadjust:
                self.tree = adjust_value(patch, io.constvar.type)
            else:
                self.tree = patch

    @classmethod
    def default(cls, type_: suite.Type, fill: str = '') -> 'ValueTreeImpl':
        default = ValueTreeImpl()
        tree = get_default_value(type_, fill)
        default.tree = tree  # type: ignore
        return default

ValueTree = Optional[ValueTreeImpl]

class SustainTree:
    """
    Value as a tree structure.

    Each node is either a literal value, as a string, or a list of nodes.

    .. Notes::

      * `'?`'` represents any sub-tree
      * `''` represents an error in the scenario or an uninitialized value
    """

    def __init__(self, type_: suite.Type):
        self.tree: CounterTree = get_default_value(type_, 0)  # type: ignore
        # number of active checks
        # self.check_count = get_type_width(type_)
        self.check_count = 0

    def patch(self, sustain: int, io: 'IO', path: SubPath):
        assert io.constvar is not None  # nosec B101  # addresses linter
        if path:
            assert isinstance(self.tree, list)  # nosec B101  # addresses linter
            offset = patch_sustain(self.tree, sustain, io.constvar.type, path)
            self.check_count += offset
        else:
            tree = get_default_value(io.constvar.type, sustain)
            self.tree = tree  # type: ignore
            self.check_count = get_type_width(io.constvar.type) if sustain > 0 else 0

    @property
    def active(self) -> bool:
        return self.check_count > 0

    def apply(self, value: ValueTreeImpl) -> bool:
        # assert self.active
        if isinstance(self.tree, list):
            assert isinstance(value.tree, list)  # nosec B101  # addresses linter
            offset = apply_sustain(value.tree, self.tree)
        else:
            offset = 0
            if self.tree >= 0:
                if self.tree == 0:
                    value.tree = '?'
                    offset = 1
                self.tree -= 1
        self.check_count -= offset
        return offset > 0


#%% classes

#{{class(63)
class ScadeProxy:
    def __init__(self):
        self.scelement: Optional[ScObject] = None

    def bind(self, model: 'Optional[ScObject]'):
        #<<132
        # abstract function
        raise NotImplementedError()
        #>>132
#}}class


#{{class(11)
class IO(ScadeProxy):
    def __init__(self, alias: str = ''):
        super().__init__()
        self.directives: List[Directive] = []
        self.sequence: Optional[Sequence] = None
        self.alias: str = alias
        #<<init
        # runtime values: reference and current values for output optimization
        self._reference: ValueTree = None
        self._value: ValueTree = None
        self._tol_reference = ''
        self._tol_value = ''
        # individual counters for checks: same tree structure as _reference/_value
        # 0: sustain just ended
        # >0 sustain active
        # -1 forever
        self._sustain: Optional[SustainTree] = None
        # number of active individual checks (scalar values)
        self._active_checks = 0
        #>>init

    @property
    def constvar(self) -> 'Optional[ScConstVar]':
        #<<99
        return self.scelement
        #>>99

    def bind(self, model: 'Optional[ScConstVar]'):
        #<<93
        self.scelement = model
        #>>93

    def is_output(self) -> bool:
        #<<151
        assert self.constvar is not None  # nosec B101  # addresses linter
        return self.constvar.is_output() or self.constvar.probe
        #>>151

    def add_directive(self, directive: 'Directive'):
        self.directives.append(directive)
        directive.io = self

    def set_sequence(self, sequence: 'Sequence'):
        self.sequence = sequence
        sequence.ios.append(self)

    #<<cls
    # properties to limit linter issues

    @property
    def value_tree(self) -> LiteralTree:
        assert self._value is not None  # nosec B101  # addresses linter
        return self._value.tree

    @property
    def reference_tree(self) -> LiteralTree:
        assert self._reference is not None  # nosec B101  # addresses linter
        return self._reference.tree
    #>>cls
#}}class


#{{class(155)
class Directive:
    def __init__(self, path: SubPath = None):
        self.step: Optional[Step] = None
        self.io: Optional[IO] = None
        self.path: SubPath = path

    def patch_io(self):
        #<<156
        # to be considered as abstract
        raise NotImplementedError()
        #>>156

    def set_step(self, step: 'Step'):
        self.step = step
        step.directives.append(self)

    def set_io(self, io: 'IO'):
        self.io = io
        io.directives.append(self)
#}}class


#{{class(36)
class SetCheck(Directive):
    def __init__(self, path: SubPath = None, value_tree: ValueTree = None):
        super().__init__(path)
        self.value_tree: ValueTree = value_tree

    def patch_io(self):
        #<<149
        assert self.value_tree is not None  # nosec B101  # addresses linter
        # assert self.path is not None  # nosec B101  # addresses linter
        assert self.io is not None  # nosec B101  # addresses linter
        assert self.io._value is not None  # nosec B101  # addresses linter
        self.io._value.patch(self.value_tree, self.io, self.path)
        #>>149
#}}class


#{{class(41)
class Check(SetCheck):
    def __init__(self, path: SubPath = None, value_tree: ValueTree = None, sustain: int = 0, tolerance: str = ''):
        super().__init__(path, value_tree)
        self.sustain: int = sustain
        self.tolerance: str = tolerance

    def patch_io(self):
        #<<150
        super().patch_io()
        assert self.io is not None  # nosec B101  # addresses linter
        assert self.io.constvar is not None  # nosec B101  # addresses linter
        assert self.io._sustain is not None  # nosec B101  # addresses linter
        self.io._sustain.patch(self.sustain, self.io, self.path)
        if self.tolerance:
            print('tolerance associated to a check not supported for', self.io.constvar.get_full_path())
        #>>150
#}}class


#{{class(39)
class Set(SetCheck):
    def __init__(self, path: SubPath = None, value_tree: ValueTree = None):
        super().__init__(path, value_tree)
#}}class


#{{class(152)
class Tolerance(Directive):
    def __init__(self, path: SubPath = None, tolerance: str = ''):
        super().__init__(path)
        self.tolerance: str = tolerance

    def patch_io(self):
        #<<153
        assert self.io is not None  # nosec B101  # addresses linter
        self.io._tol_value = self.tolerance
        #>>153
#}}class


#{{class(8)
class Step:
    def __init__(self, tolerance: str = '', cycles: int = 0):
        self.tolerance: str = tolerance
        self.comments: List[str] = []
        self.cycles: int = cycles
        self.directives: List[Directive] = []
        self.scenario: Optional[Scenario] = None

    def add_directive(self, directive: 'Directive'):
        self.directives.append(directive)
        directive.step = self

    def set_scenario(self, scenario: 'Scenario'):
        self.scenario = scenario
        scenario.steps.append(self)
#}}class


#{{class(2)
class Scenario:
    def __init__(self, pathname: str = ''):
        self.pathname: str = pathname
        self.steps: List[Step] = []
        self.sequence: Optional[Sequence] = None

    def add_step(self, step: 'Step'):
        self.steps.append(step)
        step.scenario = self

    def set_sequence(self, sequence: 'Sequence'):
        self.sequence = sequence
        sequence.scenarios.append(self)
#}}class


#{{class(7)
class CsvScenario(Scenario):
    def __init__(self, pathname: str = ''):
        super().__init__(pathname)
#}}class


#{{class(6)
class SssScenario(Scenario):
    def __init__(self, pathname: str = ''):
        super().__init__(pathname)
#}}class


#{{class(74)
class Sequence:
    def __init__(self, aliases: AliasTable = {}):
        self.application: Optional[Application] = None
        self.scenarios: List[Scenario] = []
        self.aliases: AliasTable = aliases
        self.ios: List[IO] = []
        #<<init
        # actually, implemented as Dict[str, str]
        self.aliases = {}
        # additional dict for searches (TODO: ordered dict instead?)
        self._ios = {}	# type: Dict[str, IO]
        #>>init

    def load(self) -> bool:
        #<<136
        for scenario in self.scenarios:
            loader = None
            path = Path(scenario.pathname)
            if path.suffix.lower() == '.sss':
                loader = SssLoader(self, scenario)
            elif path.suffix.lower() == '.csv':
                print('CSV scenarios not supported:', path)
                return False
            else:
                print('Unknown scenario format:', path)
                return False

            if not loader.load(path):
                return False
        return True
        #>>136

    def create_scenario(self, pathname: str) -> 'Optional[Scenario]':
        #<<137
        path = Path(pathname)
        if path.suffix.lower() == '.sss':
            scenario = SssScenario()
            scenario.pathname = pathname
            # link
            self.add_scenario(scenario)
            return scenario
        return None
        #>>137

    def find_io(self, path: str) -> 'Optional[IO]':
        #<<145
        # lazy addition of a scenario IO (either IO, sensor or probe, we don't care)
        io = self._ios.get(path)
        if not io:
            assert self.application is not None  # nosec B101  # addresses linter
            assert self.application.model is not None  # nosec B101  # addresses linter
            var = self.application.model.get_object_from_path(path)
            assert var is not None  # nosec B101  # addresses linter
            io = IO()
            io.bind(var)
            io.alias = var.get_full_path().strip('/')
            # link
            self.add_io(io)
            # cache
            self._ios[path] = io

        return io
        #>>145

    def set_application(self, application: 'Application'):
        self.application = application
        application.sequences.append(self)

    def add_scenario(self, scenario: 'Scenario'):
        self.scenarios.append(scenario)
        scenario.sequence = self

    def add_io(self, io: 'IO'):
        self.ios.append(io)
        io.sequence = self
#}}class


#{{class(22)
class Application(ScadeProxy):
    def __init__(self):
        super().__init__()
        self.sequences: List[Sequence] = []

    @property
    def model(self) -> 'Optional[ScModel]':
        #<<116
        return self.scelement
        #>>116

    def bind(self, model: 'Optional[ScModel]'):
        #<<110
        self.scelement = model
        #>>110

    def add_sequence(self, sequence: 'Sequence'):
        self.sequences.append(sequence)
        sequence.application = self
#}}class

#%% declarations

#%% end

# ---------------------------------------------------------------------------
# utilities
# ---------------------------------------------------------------------------

class SssLoader(SSSParser):
    def __init__(self, sequence: Sequence, scenario: Scenario, **kwargs):
        super().__init__(**kwargs)
        self.sequence = sequence
        self.scenario = scenario
        self.aliases = sequence.aliases

        # current step (not linked yet)
        self.cycle = Step()


    def on_cycle(self, count: str):
        # register the current cycle and initialize a new one
        self.cycle.cycles = 1 if count is None else int(count)
        # link
        self.scenario.add_step(self.cycle)
        # new step
        self.cycle = Step()


    def on_set(self, path: str, value: str):
        set_ = Set()
        path = path.strip('{}')
        path = self.aliases.get(path, path)
        set_.value_tree = ValueTreeImpl(value.strip('{}'))
        # link
        self.cycle.add_directive(set_)
        # associated variable
        path_var, items = split_path(path)
        set_.io = self.sequence.find_io(path_var)
        # sub_path
        set_.path = items


    def on_check(self, path: str, value: str, sustain: str, real: str):
        check = Check()
        path = path.strip('{}')
        path = self.aliases.get(path, path)
        check.path = path
        check.value_tree = ValueTreeImpl(value.strip('{}'))
        check.sustain = 1 if sustain is None else -1 if sustain == 'forever' else int(sustain)
        check.tolerance = real
        # link
        self.cycle.add_directive(check)
        # associated variable
        path_var, items = split_path(path)
        check.io = self.sequence.find_io(path_var)
        # sub_path
        check.path = items


    def on_uncheck(self, path: str):
        # same as checking ? forever
        path = path.strip('{}')
        path = self.aliases.get(path, path)
        self.on_check(path, '?', 'forever', '')


    def on_set_tolerance(self, path: str, real: str):
        if path:
            # associated variable
            path = path.strip('{}')
            path = self.aliases.get(path, path)
            path_var, items = split_path(path)
            if items:
                print('tolerance not supported on partial values:', path)
            else:
                tol = Tolerance()
                tol.tolerance = real
                # link
                self.cycle.add_directive(tol)
                # associated variable
                tol.io = self.sequence.find_io(path_var)
        else:
            # TODO what happens if the tolerance is set twice in the same step?
            self.cycle.tolerance = real


    def on_alias(self, alias: str, path: str):
        path = path.strip('{}')
        self.aliases[alias] = path


    def on_comment(self, line: str):
        self.cycle.comments.append(line)



# Copyright (C) 2025 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Optimization of a sss scenario to a csv scenario."""

from pathlib import Path

from scade.model.suite import Model, get_roots as get_sessions

from ansys.scade.ps.optimize_sss.scenario import (
    Application,
    Scenario,
    Sequence,
    SustainTree,
    ValueTreeImpl,
)
from ansys.scade.ps.optimize_sss.scutils import reduce_value, tree_to_str


def save_as_csv(scenario: Scenario, pathname: str):
    """Save a scenario to a CSV file."""

    def flush_line(last: bool = False):
        """Flush a line."""
        nonlocal buffer_line, buffer_cycles

        line = ';'.join(cells)
        empty_line = line.strip(';') == ''
        line_tols = ';'.join(tolerances)
        empty_tolerance = line_tols.strip(';') == ''
        if buffer_line:
            if not last and empty_tolerance and empty_line and not comments:
                # empty line: same as before
                buffer_cycles += cycles
                return

            # flush buffered line
            if last:
                f.write('Last')
                f.write(';' if buffer_cycles == 1 else ', Repeat %d;' % buffer_cycles)
            else:
                f.write(';' if buffer_cycles == 1 else 'Repeat %d;' % buffer_cycles)
            f.write(buffer_line)
            f.write('\n')

        if not last:
            for comment in comments:
                f.write('%s\n' % comment)

            if not empty_tolerance:
                f.write('Tol;')
                f.write(line_tols)
                f.write('\n')

        # buffer next line
        buffer_line = line
        buffer_cycles = cycles

    # cache for flush_line
    buffer_line = None
    buffer_cycles = 0

    # support only usual format: step by lines (format 0)
    assert scenario.sequence is not None  # nosec B101  # addresses linter
    sequence = scenario.sequence
    path = Path(pathname)
    try:
        # exist_ok added in 3.5
        path.parent.mkdir(parents=True, exist_ok=True)
    except BaseException:
        pass
    with path.open('w') as f:
        # header
        f.write('#$CsvFormat=0\n')
        # runtime values: current and reference
        for io in sequence.ios:
            assert io.constvar is not None  # nosec B101  # addresses linter
            io._reference = None
            io._value = ValueTreeImpl.default(io.constvar.type)
        # outputs: separate list with mask initialized to 0 (no check)
        os = [io for io in sequence.ios if io.is_output()]
        for o in os:
            assert io.constvar is not None  # nosec B101  # addresses linter
            o._sustain = SustainTree(io.constvar.type)
            o._active_checks = 0

        # head row
        f.write('*SCRIPT*;')
        f.write(';'.join([io.alias for io in sequence.ios]))
        f.write('\n')

        for step in scenario.steps:
            comments = step.comments

            # default values for the current step: previous ones
            for io in sequence.ios:
                assert io._value is not None  # nosec B101  # addresses linter
                io._reference = io._value
                io._value = io._reference.clone()
                io._tol_reference = io._tol_value
                # no need to clone _tol_value: str

            # patch current values with the directive
            for d in step.directives:
                d.patch_io()

            # apply active sustain
            for o in os:
                assert o._sustain is not None  # nosec B101  # addresses linter
                if o._sustain.active:
                    assert isinstance(o._value, ValueTreeImpl)  # nosec B101  # addresses linter
                    o._sustain.apply(o._value)

            # current row
            cells = [
                tree_to_str(reduce_value(io.value_tree, io.reference_tree)) for io in sequence.ios
            ]
            if step.tolerance:
                # patch tolerance for outputs w/o tolerance
                for o in os:
                    if not o._tol_value:
                        o._tol_value = step.tolerance
            tolerances = [
                '' if io._tol_value == io._tol_reference else io._tol_value for io in sequence.ios
            ]

            # split cycles if some checks become obsolete
            cycles = 1
            for i in range(step.cycles - 1):
                changed = False
                # default values for the current cycle: previous ones
                for io in sequence.ios:
                    assert io._value is not None  # nosec B101  # addresses linter
                    io._reference = io._value
                    io._value = io._reference.clone()

                for o in os:
                    assert o._sustain is not None  # nosec B101  # addresses linter
                    if o._sustain.active:
                        assert isinstance(o._value, ValueTreeImpl)  # nosec B101  # addresses linter
                        if o._sustain.apply(o._value):
                            # the current cells must be flushed
                            changed = True

                if changed:
                    # flushed buffered line
                    flush_line()

                    # current row
                    cells = [
                        tree_to_str(reduce_value(io.value_tree, io.reference_tree))
                        for io in sequence.ios
                    ]
                    # tolerance line necessary empty
                    tolerances = ['' for i in range(len(sequence.ios))]
                    # reset comments
                    comments = []
                    cycles = 1
                else:
                    cycles += 1

            flush_line()

        # loop on steps completed: flush the buffered line
        cells = 'fake'
        flush_line(last=True)


def _main(model: Model, scenario_file: str, alias_file: str, output: str) -> int:
    """Entry point for unit tests or reuse."""
    app = Application()
    app.bind(model)
    sequence = Sequence()
    app.add_sequence(sequence)
    sequence.create_scenario(alias_file)
    scenario = sequence.create_scenario(scenario_file)
    if not scenario or not sequence.load():
        print('failed to load alias/scenario files')
        return 1
    out_suffix = Path(output).suffix
    if out_suffix == '.csv':
        save_as_csv(scenario, output)
        return 0
    elif out_suffix == '.sss':
        print('sss output scenarios not supported')
        return 1
    else:
        print('unknown scenario extension', out_suffix)
        return 1


# cf __main__.py
def main(scenario_file: str, alias_file: str, output: str) -> int:
    """Entry point for ``scade.exe -script`` or called by ``__main__``."""
    model = get_sessions()[0].model
    code = _main(model, scenario_file, alias_file, output)
    return code



# Copyright (C) 2025 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Parser for SCADE Test scenarios."""

from abc import ABC, abstractmethod
from pathlib import Path
from re import compile


class SSSParser(ABC):
    """Parser abstraction for SCADE Test scenarios."""

    def __init__(self):
        self.re_cycle = compile(r'^SSM::cycle(?:\s+(\d+))?$')
        self.re_set = compile(r'^SSM::set\s+([^\s]+)\s+(.*)$')
        self.re_check = compile(
            r'^SSM::check\s+([^\s]+)\s+(.*?)(?:\s+sustain=(\w+))?(?:\s+real=(.*))?$'
        )
        self.re_uncheck = compile(r'^SSM::uncheck\s+(.*)$')
        self.re_set_tolerance = compile(r'^SSM::set_tolerance(?:\s+path=([^\s]+))?\s+real=(.*)$')
        self.re_alias = compile(r'^SSM::alias\s+(\w+)\s+(.*)$')

    # -----------------------------------------------------------------------
    # semantic actions
    # -----------------------------------------------------------------------

    @abstractmethod
    def on_cycle(self, count: str):
        """Perform semantic action for ``SSM::cycle`` directive."""
        raise NotImplementedError('Abstract method call')

    @abstractmethod
    def on_set(self, path: str, value: str):
        """Perform semantic action for ``SSM::set`` directive."""
        raise NotImplementedError('Abstract method call')

    @abstractmethod
    def on_check(self, path: str, value: str, sustain: str, real: str):
        """Perform semantic action for ``SSM::check`` directive."""
        raise NotImplementedError('Abstract method call')

    @abstractmethod
    def on_uncheck(self, path: str):
        """Perform semantic action for ``SSM::uncheck`` directive."""
        raise NotImplementedError('Abstract method call')

    @abstractmethod
    def on_set_tolerance(self, path: str, real: str):
        """Perform semantic action for ``SSM::set_tolerance`` directive."""
        raise NotImplementedError('Abstract method call')

    @abstractmethod
    def on_alias(self, alias: str, path: str):
        """Perform semantic action for ``SSM::alias`` directive."""
        raise NotImplementedError('Abstract method call')

    @abstractmethod
    def on_comment(self, line: str):
        """Register comments."""
        raise NotImplementedError('Abstract method call')

    # -----------------------------------------------------------------------
    # parsing
    # -----------------------------------------------------------------------

    def load(self, pathname: Path) -> bool:
        """Parse a SCADE Test scenario and return whether it is successful."""
        actions = {
            (self.re_set, self.on_set),
            (self.re_check, self.on_check),
            (self.re_cycle, self.on_cycle),
            (self.re_alias, self.on_alias),
            (self.re_set_tolerance, self.on_set_tolerance),
            (self.re_uncheck, self.on_uncheck),
        }

        try:
            f = pathname.open()
        except OSError as e:
            print(str(e))
            return False

        for line in f:
            line = line.strip('\n').strip()
            if len(line) == 0:
                continue

            if line[0] == '#':
                self.on_comment(line)
                continue

            for r, callback in actions:
                m = r.match(line)
                if m:
                    callback(*m.groups())
                    break
            else:
                print('syntax error:', line)
                return False

        return True
